#
#   HorusLib - Command-Line Uploader
#

# Python 3 check
import sys

if sys.version_info < (3, 6):
    print("ERROR - This script requires Python 3.6 or newer!")
    sys.exit(1)

import argparse
import codecs
import traceback
from configparser import RawConfigParser

from .habitat import *
from .decoder import decode_packet, parse_ukhas_string
from .payloads import *
from .horusudp import send_payload_summary
from .payloads import init_custom_field_list, init_payload_id_list
from .demodstats import FSKDemodStats
import horusdemodlib.payloads

def read_config(filename):
    ''' Read in the user configuation file.'''
    user_config = {
        'user_call' : 'HORUS_RX',
        'ozi_udp_port' : 55683,
        'summary_port' : 55672,
        'station_lat' : 0.0,
        'station_lon' : 0.0,
        'radio_comment' : "",
        'antenna_comment' : ""
    }

    try:
        config = RawConfigParser()
        config.read(filename)

        user_config['user_call'] = config.get('user', 'callsign')
        user_config['station_lat'] = config.getfloat('user', 'station_lat')
        user_config['station_lon'] = config.getfloat('user', 'station_lon')
        user_config['radio_comment'] = config.get('user', 'radio_comment')
        user_config['antenna_comment'] = config.get('user', 'antenna_comment')
        user_config['ozi_udp_port'] = config.getint('horus_udp', 'ozimux_port')
        user_config['summary_port'] = config.getint('horus_udp', 'summary_port')

        return user_config

    except:
        traceback.print_exc()
        logging.error("Could not parse config file, exiting. Have you copied user.cfg.example to user.cfg?")
        return None


def main():

    # Read command-line arguments
    parser = argparse.ArgumentParser(description="Project Horus Binary/RTTY Telemetry Handler", formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('-c', '--config', type=str, default='user.cfg', help="Configuration file to use. Default: user.cfg")
    parser.add_argument("--noupload", action="store_true", default=False, help="Disable Habitat upload.")
    parser.add_argument("--rtty", action="store_true", default=False, help="Expect only RTTY inputs, do not update payload lists.")
    parser.add_argument("--log", type=str, default="telemetry.log", help="Write decoded telemetry to this log file.")
    parser.add_argument("--debuglog", type=str, default="horusb_debug.log", help="Write debug log to this file.")
    parser.add_argument("--payload-list", type=str, default="payload_id_list.txt", help="List of known payload IDs.")
    parser.add_argument("--custom-fields", type=str, default="custom_field_list.json", help="List of payload Custom Fields")
    parser.add_argument("--nodownload", action="store_true", default=False, help="Do not download new lists.")
#   parser.add_argument("--ozimux", type=int, default=-1, help="Override user.cfg OziMux output UDP port. (NOT IMPLEMENTED)")
#   parser.add_argument("--summary", type=int, default=-1, help="Override user.cfg UDP Summary output port. (NOT IMPLEMENTED)")
    parser.add_argument("-v", "--verbose", action="store_true", default=False, help="Verbose output (set logging level to DEBUG)")
    args = parser.parse_args()

    if args.verbose:
        logging_level = logging.DEBUG
    else:
        logging_level = logging.INFO

    # Set up logging
    logging.basicConfig(format="%(asctime)s %(levelname)s: %(message)s", level=logging_level)

    # Read in the configuration file.
    user_config = read_config(args.config)

    # If we could not read the configuration file, exit.
    if user_config == None:
        logging.critical(f"Could not load {args.config}, exiting...")
        sys.exit(1)

    if args.log != "none":
        _logfile = open(args.log, 'a')
        logging.info(f"Opened log file {args.log}.")
    else:
        _logfile = None


    if args.rtty == False:

        if args.nodownload:
            logging.info("Using local lists.")
            horusdemodlib.payloads.HORUS_PAYLOAD_LIST = read_payload_list(filename=args.payload_list)
            horusdemodlib.payloads.HORUS_CUSTOM_FIELDS = read_custom_field_list(filename=args.custom_fields)
        else:
            # Downlaod
            horusdemodlib.payloads.HORUS_PAYLOAD_LIST = init_payload_id_list(filename=args.payload_list)
            horusdemodlib.payloads.HORUS_CUSTOM_FIELDS = init_custom_field_list(filename=args.custom_fields)
            
        logging.info(f"Payload list contains {len(list(horusdemodlib.payloads.HORUS_PAYLOAD_LIST.keys()))} entries.")

        logging.info(f"Custom Field list contains {len(list(horusdemodlib.payloads.HORUS_CUSTOM_FIELDS.keys()))} entries.")

    # Start the Habitat uploader thread.
    habitat_uploader = HabitatUploader(
        user_callsign = user_config['user_call'],
        listener_lat = user_config['station_lat'],
        listener_lon = user_config['station_lon'],
        listener_radio = user_config['radio_comment'],
        listener_antenna = user_config['antenna_comment'],
        inhibit=args.noupload
    )

    logging.info("Using User Callsign: %s" % user_config['user_call'])

    demod_stats = FSKDemodStats()

    logging.info("Started Horus Demod Uploader. Hit CTRL-C to exit.")
    # Main loop
    try:
        while True:
            # Read lines in from stdin, and strip off any trailing newlines
            data = sys.stdin.readline()

            if (data == ''):
                # Empty line means stdin has been closed.
                logging.info("Caught EOF, exiting.")
                break

            # Otherwise, strip any newlines, and continue.
            data = data.rstrip()

            # If the line of data starts with '$$', we assume it is a UKHAS-standard ASCII telemetry sentence.
            # Otherwise, we assume it is a string of hexadecimal bytes, and attempt to parse it as a binary telemetry packet.

            if data.startswith('$$'):
                # RTTY packet handling.
                # Attempt to extract fields from it:
                logging.info(f"Received raw RTTY packet: {data}")
                try:
                    _decoded = parse_ukhas_string(data)
                    # If we get here, the string is valid!

                    # Add in SNR data.
                    _snr = demod_stats.snr
                    _decoded['snr'] = _snr

                    # Send via UDP
                    send_payload_summary(_decoded, port=user_config['summary_port'])

                    # Upload the string to Habitat
                    _decoded_str = "$$" + data.split('$')[-1] + '\n'
                    habitat_uploader.add(_decoded_str)

                    if _logfile:
                        _logfile.write(_decoded_str)
                        _logfile.flush()

                    logging.info(f"Decoded String (SNR {demod_stats.snr:.1f} dB): {_decoded_str[:-1]}")

                except Exception as e:
                    logging.error(f"Decode Failed: {str(e)}")
            
            elif data.startswith('{'):
                # Possibly a line of modem statistics, attempt to decode it.
                demod_stats.update(data)

            else:
                # Handle binary packets
                logging.info(f"Received raw binary packet: {data}")
                try:
                    _binary_string = codecs.decode(data, 'hex')
                except TypeError as e:
                    logging.error("Error parsing line as hexadecimal (%s): %s" % (str(e), data))
                    continue

                try:
                    _decoded = decode_packet(_binary_string)
                    # If we get here, we have a valid packet!
                    
                    # Add in SNR data.
                    _snr = demod_stats.snr
                    _decoded['snr'] = _snr

                    # Send via UDP
                    send_payload_summary(_decoded, port=user_config['summary_port'])

                    # Upload to Habitat
                    habitat_uploader.add(_decoded['ukhas_str']+'\n')

                    if _logfile:
                        _logfile.write(_decoded['ukhas_str']+'\n')
                        _logfile.flush()

                    logging.info(f"Decoded Binary Packet (SNR {demod_stats.snr:.1f} dB): {_decoded['ukhas_str']}")
                except Exception as e:
                    logging.error(f"Decode Failed: {str(e)}")

    except KeyboardInterrupt:
        logging.info("Caught CTRL-C, exiting.")

    habitat_uploader.close()

if __name__ == "__main__":
    main()


#!/usr/bin/env python
#
#   Horus Demod Library - Habitat Uploader
#
#   Mark Jessop <vk5qi@rfhead.net>
#

import datetime
import json
import logging
import random
import requests
import time
from base64 import b64encode
from hashlib import sha256
from queue import Queue
from threading import Thread


class HabitatUploader(object):
    """ 
    Queued Habitat Telemetry Uploader class 
    
    Packets to be uploaded to Habitat are added to a queue for uploading.
    If an upload attempt times out, the packet is discarded.
    If the queue fills up (probably indicating no network connection, and a fast packet downlink rate),
    it is immediately emptied, to avoid upload of out-of-date packets.
    """

    HABITAT_URL = "http://habitat.habhub.org/"
    HABITAT_DB = "habitat"
    HABITAT_UUIDS = HABITAT_URL + "_uuids?count=%d"
    HABITAT_DB_URL = HABITAT_URL + HABITAT_DB + "/"

    def __init__(
        self,
        user_callsign="FSK_DEMOD",
        listener_lat=0.0,
        listener_lon=0.0,
        listener_radio="",
        listener_antenna="",
        queue_size=64,
        upload_timeout=10,
        upload_retries=5,
        upload_retry_interval=0.25,
        inhibit=False,
    ):
        """ Create a Habitat Uploader object. """

        self.upload_timeout = upload_timeout
        self.upload_retries = upload_retries
        self.upload_retry_interval = upload_retry_interval
        self.queue_size = queue_size
        self.habitat_upload_queue = Queue(queue_size)
        self.inhibit = inhibit

        # Listener information
        self.user_callsign = user_callsign
        self.listener_lat = listener_lat
        self.listener_lon = listener_lon
        self.listener_radio = listener_radio
        self.listener_antenna = listener_antenna
        self.position_uploaded = False

        self.callsign_init = False
        self.uuids = []

        if self.inhibit:
            logging.info("Habitat Uploader inhibited.")

        # Start the uploader thread.
        self.habitat_uploader_running = True
        self.uploadthread = Thread(target=self.habitat_upload_thread)
        self.uploadthread.start()

    def habitat_upload(self, sentence):
        """ Upload a UKHAS-standard telemetry sentence to Habitat """

        # Generate payload to be uploaded
        # b64encode accepts and returns bytes objects.
        _sentence_b64 = b64encode(sentence.encode("ascii"))
        _date = datetime.datetime.utcnow().isoformat("T") + "Z"
        _user_call = self.user_callsign

        _data = {
            "type": "payload_telemetry",
            "data": {
                "_raw": _sentence_b64.decode(
                    "ascii"
                )  # Convert back to a string to be serialisable
            },
            "receivers": {
                _user_call: {"time_created": _date, "time_uploaded": _date,},
            },
        }

        # The URl to upload to.
        _url = f"{self.HABITAT_URL}{self.HABITAT_DB}/_design/payload_telemetry/_update/add_listener/{sha256(_sentence_b64).hexdigest()}"

        # Delay for a random amount of time between 0 and upload_retry_interval*2 seconds.
        time.sleep(random.random() * self.upload_retry_interval * 2.0)

        _retries = 0

        # When uploading, we have three possible outcomes:
        # - Can't connect. No point re-trying in this situation.
        # - The packet is uploaded successfult (201 / 403)
        # - There is a upload conflict on the Habitat DB end (409). We can retry and it might work.
        while _retries < self.upload_retries:
            # Run the request.
            try:
                _req = requests.put(
                    _url, data=json.dumps(_data), timeout=self.upload_timeout
                )
            except Exception as e:
                logging.error("Habitat - Upload Failed: %s" % str(e))
                break

            if _req.status_code == 201 or _req.status_code == 403:
                # 201 = Success, 403 = Success, sentence has already seen by others.
                logging.info(f"Habitat - Uploaded sentence: {sentence.strip()}")
                _upload_success = True
                break
            elif _req.status_code == 409:
                # 409 = Upload conflict (server busy). Sleep for a moment, then retry.
                logging.debug("Habitat - Upload conflict.. retrying.")
                time.sleep(random.random() * self.upload_retry_interval)
                _retries += 1
            else:
                logging.error(
                    "Habitat - Error uploading to Habitat. Status Code: %d."
                    % _req.status_code
                )
                break

        if _retries == self.upload_retries:
            logging.error(
                "Habitat - Upload conflict not resolved with %d retries."
                % self.upload_retries
            )

        return

    def habitat_upload_thread(self):
        """ Handle uploading of packets to Habitat """

        logging.info("Started Habitat Uploader Thread.")

        while self.habitat_uploader_running:

            if self.habitat_upload_queue.qsize() > 0:
                # If the queue is completely full, jump to the most recent telemetry sentence.
                if self.habitat_upload_queue.qsize() == self.queue_size:
                    while not self.habitat_upload_queue.empty():
                        sentence = self.habitat_upload_queue.get()

                    logging.warning(
                        "Habitat uploader queue was full - possible connectivity issue."
                    )
                else:
                    # Otherwise, get the first item in the queue.
                    sentence = self.habitat_upload_queue.get()

                # Attempt to upload it.
                self.habitat_upload(sentence)

            else:
                # Wait for a short time before checking the queue again.
                time.sleep(0.5)

            if not self.position_uploaded:
                # Validate the lat/lon entries.
                try:
                    _lat = float(self.listener_lat)
                    _lon = float(self.listener_lon)

                    if (_lat != 0.0) or (_lon != 0.0):
                        _success = self.uploadListenerPosition(
                            self.user_callsign,
                            _lat,
                            _lon,
                            self.listener_radio,
                            self.listener_antenna,
                        )
                    else:
                        logging.warning("Listener position set to 0.0/0.0 - not uploading.")
                
                except Exception as e:
                    logging.error("Error uploading listener position: %s" % str(e))

                # Set this flag regardless if the upload worked.
                # The user can trigger a re-upload.
                self.position_uploaded = True


        logging.info("Stopped Habitat Uploader Thread.")

    def add(self, sentence):
        """ Add a sentence to the upload queue """

        if self.inhibit:
            # We have upload inhibited. Return.
            return

        # Handling of arbitrary numbers of $$'s at the start of a sentence:
        # Extract the data part of the sentence (i.e. everything after the $$'s')
        sentence = sentence.split("$")[-1]
        # Now add the *correct* number of $$s back on.
        sentence = "$$" + sentence

        if not (sentence[-1] == "\n"):
            sentence += "\n"

        try:
            self.habitat_upload_queue.put_nowait(sentence)
        except Exception as e:
            logging.error("Error adding sentence to queue: %s" % str(e))

    def close(self):
        """ Shutdown uploader thread. """
        self.habitat_uploader_running = False

    def ISOStringNow(self):
        return "%sZ" % datetime.datetime.utcnow().isoformat()

    def postListenerData(self, doc, timeout=10):

        # do we have at least one uuid, if not go get more
        if len(self.uuids) < 1:
            self.fetchUuids()

        # Attempt to add UUID and time data to document.
        try:
            doc["_id"] = self.uuids.pop()
        except IndexError:
            logging.error(
                "Habitat - Unable to post listener data - no UUIDs available."
            )
            return False

        doc["time_uploaded"] = self.ISOStringNow()

        try:
            _r = requests.post(
                f"{self.HABITAT_URL}{self.HABITAT_DB}/", json=doc, timeout=timeout
            )
            return True
        except Exception as e:
            logging.error("Habitat - Could not post listener data - %s" % str(e))
            return False

    def fetchUuids(self, timeout=10):

        _retries = 5

        while _retries > 0:
            try:
                _r = requests.get(self.HABITAT_UUIDS % 10, timeout=timeout)
                self.uuids.extend(_r.json()["uuids"])
                logging.debug("Habitat - Got UUIDs")
                return
            except Exception as e:
                logging.error(
                    "Habitat - Unable to fetch UUIDs, retrying in 2 seconds - %s"
                    % str(e)
                )
                time.sleep(2)
                _retries = _retries - 1
                continue

        logging.error("Habitat - Gave up trying to get UUIDs.")
        return

    def initListenerCallsign(self, callsign, radio="", antenna=""):
        doc = {
            "type": "listener_information",
            "time_created": self.ISOStringNow(),
            "data": {"callsign": callsign, "antenna": antenna, "radio": radio,},
        }

        resp = self.postListenerData(doc)

        if resp is True:
            logging.debug("Habitat - Listener Callsign Initialized.")
            return True
        else:
            logging.error("Habitat - Unable to initialize callsign.")
            return False

    def uploadListenerPosition(self, callsign, lat, lon, radio="", antenna=""):
        """ Initializer Listener Callsign, and upload Listener Position """

        # Attempt to initialize the listeners callsign
        resp = self.initListenerCallsign(callsign, radio=radio, antenna=antenna)
        # If this fails, it means we can't contact the Habitat server,
        # so there is no point continuing.
        if resp is False:
            return False

        doc = {
            "type": "listener_telemetry",
            "time_created": self.ISOStringNow(),
            "data": {
                "callsign": callsign,
                "chase": False,
                "latitude": lat,
                "longitude": lon,
                "altitude": 0,
                "speed": 0,
            },
        }

        # post position to habitat
        resp = self.postListenerData(doc)
        if resp is True:
            logging.info("Habitat - Listener information uploaded.")
            return True
        else:
            logging.error("Habitat - Unable to upload listener information.")
            return False

    def trigger_position_upload(self):
        """ Trigger a re-upload of the listener position """
        self.position_uploaded = False


if __name__ == "__main__":

    # Setup Logging
    logging.basicConfig(
        format="%(asctime)s %(levelname)s: %(message)s", level=logging.INFO
    )

    habitat = HabitatUploader(
        user_callsign="HORUSGUI_TEST",
        listener_lat=-34.0,
        listener_lon=138.0,
        listener_radio="Testing Habitat Uploader",
        listener_antenna="Wet Noodle",
    )

    habitat.add("$$DUMMY,0,0.0,0.0*F000")

    time.sleep(10)
    habitat.trigger_position_upload()
    time.sleep(5)
    habitat.close()
